"""Preprocessing utilities

In order to preprocess ZIP file to extract a depthmap, we use this code:
https://github.com/Welthungerhilfe/cgm-rg/blob/92efa0febb91c9656ce8e5dbfad953ff7ce721a9/src/utils/preprocessing.py#L12

file of minor importance:
https://github.com/Welthungerhilfe/cgm-ml/blob/c8be9138e025845bedbe7cfc0d131ef668e01d4b/old
/cgm_database/command_preprocess.py#L92
"""

from pathlib import Path
import pickle
from typing import Dict, List, Tuple
import zipfile

import numpy as np
from skimage.transform import resize

NORMALIZATION_VALUE = 7.5
IMAGE_TARGET_HEIGHT, IMAGE_TARGET_WIDTH = 180, 240


def load_depth(fpath: str) -> Tuple[bytes, int, int, float, float]:
    """Take ZIP file and extract depth and metadata
    Args:
        fpath (str): File path to the ZIP
    Returns:
        depth_data (bytes): depthmap data
        width(int): depthmap width in pixel
        height(int): depthmap height in pixel
        depth_scale(float)
        max_confidence(float)
    """

    with zipfile.ZipFile(fpath) as z:
        with z.open('data') as f:
            # Example for a first_line:
            #     '180x135_0.001_7_0.57045287_-0.0057296_0.0022602521_0.82130724_-0.059177425_0.0024800065_0.030834956'
            first_line = f.readline().decode().strip()

            file_header = first_line.split("_")

            # header[0] example: 180x135
            width, height = file_header[0].split("x")
            width, height = int(width), int(height)
            depth_scale = float(file_header[1])
            max_confidence = float(file_header[2])

            depth_data = f.read()
    return depth_data, width, height, depth_scale, max_confidence


def parse_depth(tx: int, ty: int, data: bytes, depth_scale: float, width: int) -> float:
    assert isinstance(tx, int)
    assert isinstance(ty, int)

    depth = data[(ty * width + tx) * 3 + 0] << 8
    depth += data[(ty * width + tx) * 3 + 1]

    depth *= depth_scale
    return depth


def preprocess_depthmap(depthmap: np.array) -> np.array:
    return depthmap.astype("float32")


def preprocess(depthmap: np.array) -> np.array:
    depthmap = preprocess_depthmap(depthmap)
    depthmap = depthmap / NORMALIZATION_VALUE
    depthmap = resize(depthmap, (IMAGE_TARGET_HEIGHT, IMAGE_TARGET_WIDTH))
    depthmap = depthmap.reshape((depthmap.shape[0], depthmap.shape[1], 1))
    return depthmap


def prepare_depthmap(data: bytes, width: int, height: int, depth_scale: float) -> np.array:
    """Convert bytes array into np.array"""
    output = np.zeros((width, height, 1))
    for cx in range(width):
        for cy in range(height):
            # depth data scaled to be visible
            output[cx][height - cy - 1] = parse_depth(cx, cy, data, depth_scale, width)
    arr = np.array(output, dtype='float32')
    return arr.reshape(width, height)


def get_depthmaps(fpaths: List[str]) -> np.array:
    depthmaps = []
    for fpath in fpaths:
        data, width, height, depthScale, _ = load_depth(fpath)
        depthmap = prepare_depthmap(data, width, height, depthScale)
        depthmap = preprocess(depthmap)
        depthmaps.append(depthmap)

    depthmaps = np.array(depthmaps)
    return depthmaps


class ArtifactProcessor:
    def __init__(self, input_dir: str, output_dir: str, idx2col: Dict[int, str]):
        self.input_dir = input_dir
        self.output_dir = output_dir
        self.idx2col = idx2col

    def create_and_save_pickle(self,
                               zip_input_full_path: str,
                               timestamp: str,
                               scan_id: str,
                               scan_step: str,
                               target_tuple: Tuple[float, float, float],
                               order_number: str) -> str:
        """Side effect: Saves and returns file path"""
        depthmaps = get_depthmaps([zip_input_full_path])
        pickle_output_path = f"qrcode/{scan_id}/{scan_step}/pc_{scan_id}_{timestamp}_{scan_step}_{order_number}.p"
        pickle_output_full_path = f"{self.output_dir}/{pickle_output_path}"
        Path(pickle_output_full_path).parent.mkdir(parents=True, exist_ok=True)
        pickle.dump((depthmaps, np.array(target_tuple)), open(pickle_output_full_path, "wb"))
        return pickle_output_full_path

    def process_artifact_tuple(self, artifact_tuple: tuple):
        """Side effect: Saves and returns file path"""
        artifact_dict = {self.idx2col[i]: el for i, el in enumerate(artifact_tuple)}
        target_tuple = (artifact_dict['height'], artifact_dict['weight'], artifact_dict['muac'])
        zip_input_full_path = f"{self.input_dir}/{artifact_dict['file_path']}"

        pickle_output_full_path = self.create_and_save_pickle(
            zip_input_full_path=zip_input_full_path,
            timestamp=artifact_dict['timestamp'],
            scan_id=artifact_dict['scan_id'],
            scan_step=artifact_dict['scan_step'],
            target_tuple=target_tuple,
            order_number=artifact_dict['order_number'],
        )
        return pickle_output_full_path



import logging
import logging.config

import matplotlib.pyplot as plt
import numpy as np

from constants import MASK_CHILD
from depthmap import Depthmap

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d')

PATTERN_LENGTH_IN_METERS = 0.1

SUBPLOT_DEPTH = 0
SUBPLOT_NORMAL = 1
SUBPLOT_SEGMENTATION = 2
SUBPLOT_CONFIDENCE = 3
SUBPLOT_RGB = 4
SUBPLOT_COUNT = 5


def render_pixel(output: object,
                 x: int,
                 y: int,
                 floor: float,
                 mask: np.array,
                 dmap: Depthmap):

    depth = dmap.parse_depth(x, y)
    if not depth:
        return

    # convert ToF coordinates into RGB coordinates
    vec = dmap.convert_2d_to_3d(1, x, y, depth)
    vec = dmap.convert_3d_to_2d(0, vec[0], vec[1], vec[2])

    # depth data visualisation (scaled to be visible)
    index = SUBPLOT_DEPTH * dmap.height + dmap.height - y - 1
    output[x][index] = 1.0 - min(depth / 2.0, 1.0)

    # normal vector visualisation
    normal = dmap.calculate_normal_vector(x, y)
    index = SUBPLOT_NORMAL * dmap.height + dmap.height - y - 1
    output[x][index][0] = abs(normal[0])
    output[x][index][1] = abs(normal[1])
    output[x][index][2] = abs(normal[2])

    # segmentation visualisation
    point = dmap.convert_2d_to_3d_oriented(1, x, y, depth)
    horizontal = (point[1] % PATTERN_LENGTH_IN_METERS) / PATTERN_LENGTH_IN_METERS
    vertical_x = (point[0] % PATTERN_LENGTH_IN_METERS) / PATTERN_LENGTH_IN_METERS
    vertical_z = (point[2] % PATTERN_LENGTH_IN_METERS) / PATTERN_LENGTH_IN_METERS
    vertical = (vertical_x + vertical_z) / 2.0
    index = SUBPLOT_SEGMENTATION * dmap.height + dmap.height - y - 1
    if mask[x][y] == MASK_CHILD:
        output[x][index][0] = horizontal / (depth * depth)
        output[x][index][1] = horizontal / (depth * depth)
    elif abs(normal[1]) < 0.5:
        output[x][index][0] = horizontal / (depth * depth)
    elif abs(normal[1]) > 0.5:
        if abs(point[1] - floor) < 0.1:
            output[x][index][2] = vertical / (depth * depth)
        else:
            output[x][index][1] = vertical / (depth * depth)

    # confidence value visualisation
    index = SUBPLOT_CONFIDENCE * dmap.height + dmap.height - y - 1
    output[x][index][:] = dmap.parse_confidence(x, y)
    if output[x][index][0] == 0:
        output[x][index][:] = 1

    # RGB data visualisation
    index = SUBPLOT_RGB * dmap.height + dmap.height - y - 1
    if 0 < vec[0] < dmap.width and 1 < vec[1] < dmap.height and dmap.has_rgb:
        output[x][index][0] = dmap.rgb_array[int(vec[1])][int(vec[0])][0] / 255.0
        output[x][index][1] = dmap.rgb_array[int(vec[1])][int(vec[0])][1] / 255.0
        output[x][index][2] = dmap.rgb_array[int(vec[1])][int(vec[0])][2] / 255.0

    # ensure pixel clipping
    for i in range(SUBPLOT_COUNT):
        index = i * dmap.height + dmap.height - y - 1
        output[x][index][0] = min(max(0, output[x][index][0]), 1)
        output[x][index][1] = min(max(0, output[x][index][1]), 1)
        output[x][index][2] = min(max(0, output[x][index][2]), 1)


def render_plot(dmap: Depthmap):
    # floor and child detection
    floor = dmap.get_floor_level()
    mask, highest = dmap.detect_child(floor)

    # render the visualisations
    output = np.zeros((dmap.width, dmap.height * SUBPLOT_COUNT, 3))
    for x in range(dmap.width):
        for y in range(dmap.height):
            render_pixel(output, x, y, floor, mask, dmap)

    logging.info('height=%fm', highest - floor)
    plt.imshow(output)



import argparse
import logging
import logging.config
import os
import pickle
import random
import shutil
import time
from importlib import import_module
from pathlib import Path
from typing import List

import glob2 as glob
import numpy as np
import pandas as pd
import tensorflow as tf
from azureml.core import Experiment, Workspace
from azureml.core.run import Run
from tensorflow.keras.models import load_model

from constants import DATA_DIR_ONLINE_RUN, DEFAULT_CONFIG, REPO_DIR


def copy_dir(src: Path, tgt: Path, glob_pattern: str, should_touch_init: bool = False):
    logging.info("Creating temp folder")
    if tgt.exists():
        shutil.rmtree(tgt)
    tgt.mkdir(parents=True, exist_ok=True)
    if should_touch_init:
        (tgt / '__init__.py').touch(exist_ok=False)

    paths_to_copy = list(src.glob(glob_pattern))
    logging.info(f"Copying to {tgt} the following files: {str(paths_to_copy)}")
    for p in paths_to_copy:
        destpath = tgt / p.relative_to(src)
        destpath.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(p, destpath)


# Get the current run.
run = Run.get_context()

if run.id.startswith("OfflineRun"):

    # Copy common into the temp folder
    common_dir_path = REPO_DIR / "src/common"
    temp_common_dir = Path(__file__).parent / "temp_common"
    copy_dir(src=common_dir_path, tgt=temp_common_dir, glob_pattern='*/*.py', should_touch_init=True)

from temp_common.evaluation.constants_eval import (  # noqa: E402, F401
    AGE_IDX, COLUMN_NAME_AGE, COLUMN_NAME_GOODBAD, COLUMN_NAME_SEX,
    GOODBAD_DICT, GOODBAD_IDX, HEIGHT_IDX, SEX_IDX, WEIGHT_IDX)
from temp_common.evaluation.eval_utils import (  # noqa: E402, F401
    avgerror, calculate_performance, extract_qrcode,
    extract_scantype, preprocess_depthmap, preprocess_targets
)
from temp_common.evaluation.eval_utilities import (  # noqa: E402, F401
    calculate_and_save_results,
    download_model,
    calculate_performance_age, calculate_performance_goodbad,
    calculate_performance_sex, download_dataset, draw_age_scatterplot,
    draw_stunting_diagnosis, draw_uncertainty_goodbad_plot,
    draw_uncertainty_scatterplot, draw_wasting_diagnosis, filter_dataset_according_to_standing_lying,
    get_column_list, get_dataset_path,
    get_depthmap_files, get_model_path)
from temp_common.evaluation.uncertainty_utils import \
    get_prediction_uncertainty_deepensemble  # noqa: E402, F401
from temp_common.model_utils.preprocessing_multiartifact_python import \
    create_multiartifact_paths_for_qrcodes  # noqa: E402, F401
from temp_common.model_utils.preprocessing_multiartifact_tensorflow import \
    create_multiartifact_sample  # noqa: E402, F401


logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d')

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--qa_config_module", default=DEFAULT_CONFIG, help="Configuration file")
    args = parser.parse_args()
    qa_config_module = args.qa_config_module
    qa_config = import_module(qa_config_module)
else:
    qa_config_module = DEFAULT_CONFIG
    qa_config = import_module(qa_config_module)
logging.info('Using the following config: %s', qa_config_module)


MODEL_CONFIG = qa_config.MODEL_CONFIG
EVAL_CONFIG = qa_config.EVAL_CONFIG
DATA_CONFIG = qa_config.DATA_CONFIG
RESULT_CONFIG = qa_config.RESULT_CONFIG
FILTER_CONFIG = qa_config.FILTER_CONFIG if getattr(qa_config, 'FILTER_CONFIG', False) else None


RUN_ID = MODEL_CONFIG.RUN_ID if getattr(MODEL_CONFIG, 'RUN_ID', False) else None
RUN_IDS = MODEL_CONFIG.RUN_IDS if getattr(MODEL_CONFIG, 'RUN_IDS', False) else None
assert bool(RUN_ID) != bool(RUN_IDS), 'RUN_ID xor RUN_IDS needs to be defined'

# Function for loading and processing depthmaps.


def tf_load_pickle(path, max_value):
    """Utility to load the depthmap pickle file"""
    def py_load_pickle(path, max_value):
        if FILTER_CONFIG is not None:
            depthmap, targets, _image = pickle.load(open(path.numpy(), "rb"))  # for filter (Contains RGBs)
        else:
            depthmap, targets = pickle.load(open(path.numpy(), "rb"))
        depthmap = preprocess_depthmap(depthmap)
        depthmap = depthmap / max_value
        depthmap = tf.image.resize(depthmap, (DATA_CONFIG.IMAGE_TARGET_HEIGHT, DATA_CONFIG.IMAGE_TARGET_WIDTH))
        targets = preprocess_targets(targets, DATA_CONFIG.TARGET_INDEXES)
        return depthmap, targets

    depthmap, targets = tf.py_function(py_load_pickle, [path, max_value], [tf.float32, tf.float32])
    depthmap.set_shape((DATA_CONFIG.IMAGE_TARGET_HEIGHT, DATA_CONFIG.IMAGE_TARGET_WIDTH, 1))
    targets.set_shape((len(DATA_CONFIG.TARGET_INDEXES,)))
    return path, depthmap, targets


def prepare_sample_dataset(df_sample, dataset_path):
    df_sample['artifact_path'] = df_sample.apply(
        lambda x: f"{dataset_path}/{x['qrcode']}/{x['scantype']}/{x['artifact']}", axis=1)
    paths_evaluation = list(df_sample['artifact_path'])
    dataset_sample = tf.data.Dataset.from_tensor_slices(paths_evaluation)
    dataset_sample = dataset_sample.map(lambda path: tf_load_pickle(path, DATA_CONFIG.NORMALIZATION_VALUE))
    dataset_sample = dataset_sample.map(lambda _path, depthmap, targets: (depthmap, targets))
    dataset_sample = dataset_sample.cache()
    dataset_sample = dataset_sample.prefetch(tf.data.experimental.AUTOTUNE)
    return dataset_sample


def get_prediction_multiartifact(model_path: str, samples_paths: List[List[str]]) -> List[List[str]]:
    """Make prediction on each multiartifact sample.

    Args:
        model_path: File path to the model
        samples_paths: A list of samples where each sample contains N_ARTIFACTS.

    Returns:
        List with tuples: ('artifacts', 'predicted', 'GT')
    """
    logging.info("loading model from %s", model_path)
    model = load_model(model_path, compile=False)

    predictions = []
    for sample_paths in samples_paths:
        depthmap, targets = create_multiartifact_sample(sample_paths,
                                                        DATA_CONFIG.NORMALIZATION_VALUE,
                                                        DATA_CONFIG.IMAGE_TARGET_HEIGHT,
                                                        DATA_CONFIG.IMAGE_TARGET_WIDTH,
                                                        tf.constant(DATA_CONFIG.TARGET_INDEXES),
                                                        DATA_CONFIG.N_ARTIFACTS)
        depthmaps = tf.stack([depthmap])
        pred = model.predict(depthmaps)
        predictions.append([sample_paths[0], float(np.squeeze(pred)), targets[0]])
    return predictions


def get_prediction(model_path: str, dataset_evaluation: tf.data.Dataset) -> np.array:
    """Perform the prediction on the dataset with the given model.

    Args:
        model_path: Path of the trained model
        dataset_evaluation: dataset in which the evaluation need to performed
    Returns:
        predictions, array shape (N_SAMPLES, )
    """
    logging.info("loading model from %s", model_path)
    model = load_model(model_path, compile=False)

    dataset = dataset_evaluation.batch(DATA_CONFIG.BATCH_SIZE)

    logging.info("starting predicting")
    start = time.time()
    predictions = model.predict(dataset, batch_size=DATA_CONFIG.BATCH_SIZE)
    end = time.time()
    logging.info("Total time for uncertainty prediction experiment: %.2f sec", end - start)

    prediction_list = np.squeeze(predictions)
    return prediction_list


def get_predictions_from_multiple_models(model_paths: list, dataset_evaluation: tf.data.Dataset) -> list:
    prediction_list_one = []
    for model_index, model_path in enumerate(model_paths):
        logging.info(f"Model {model_index + 1}/{len(model_paths)}")
        prediction_list_one += [get_prediction(model_path, dataset_evaluation)]
        logging.info("Prediction made by model on the depthmaps...")
    prediction_list_one = np.array(prediction_list_one)
    prediction_list_one = np.mean(prediction_list_one, axis=0)
    return prediction_list_one


if __name__ == "__main__":

    # Make experiment reproducible
    tf.random.set_seed(EVAL_CONFIG.SPLIT_SEED)
    random.seed(EVAL_CONFIG.SPLIT_SEED)

    OUTPUT_CSV_PATH = str(REPO_DIR / 'data'
                          / RESULT_CONFIG.SAVE_PATH) if run.id.startswith("OfflineRun") else RESULT_CONFIG.SAVE_PATH
    if RUN_ID is not None:
        MODEL_BASE_DIR = REPO_DIR / 'data' / RUN_ID if run.id.startswith("OfflineRun") else Path('.')
    if RUN_IDS is not None:
        MODEL_BASE_DIR = REPO_DIR / 'data' / \
            MODEL_CONFIG.EXPERIMENT_NAME if run.id.startswith("OfflineRun") else Path('.')

    # Offline run. Download the sample dataset and run locally. Still push results to Azure.
    if run.id.startswith("OfflineRun"):
        logging.info("Running in offline mode...")

        # Access workspace.
        logging.info("Accessing workspace...")
        workspace = Workspace.from_config()
        experiment = Experiment(workspace, EVAL_CONFIG.EXPERIMENT_NAME)
        run = experiment.start_logging(outputs=None, snapshot_directory=None)

        # Get dataset.
        logging.info("Accessing dataset...")
        dataset_name = DATA_CONFIG.NAME
        dataset_path = str(REPO_DIR / "data" / dataset_name)
        if not os.path.exists(dataset_path):
            dataset = workspace.datasets[dataset_name]
            dataset.download(target_path=dataset_path, overwrite=False)

    # Online run. Use dataset provided by training notebook.
    else:
        logging.info("Running in online mode...")
        experiment = run.experiment
        workspace = experiment.workspace
        dataset_name = DATA_CONFIG.NAME

        # Download
        dataset_path = get_dataset_path(DATA_DIR_ONLINE_RUN, dataset_name)
        download_dataset(workspace, dataset_name, dataset_path)

    if RUN_IDS is not None:
        for run_id in RUN_IDS:
            logging.info(f"Downloading run {run_id}")
            download_model(
                workspace=workspace,
                experiment_name=MODEL_CONFIG.EXPERIMENT_NAME,
                run_id=run_id,
                input_location=os.path.join(MODEL_CONFIG.INPUT_LOCATION, MODEL_CONFIG.NAME),
                output_location=MODEL_BASE_DIR / run_id
            )

        model_paths = glob.glob(os.path.join(MODEL_BASE_DIR, "*"))
        model_paths = [path for path in model_paths if os.path.isdir(path)]
        model_paths = [path for path in model_paths if path.split("/")[-1].startswith(MODEL_CONFIG.EXPERIMENT_NAME)]
        model_paths = [os.path.join(path, "outputs", "best_model.ckpt") for path in model_paths]
        logging.info(f"Models paths ({len(model_paths)}):")
        logging.info("\t" + "\n\t".join(model_paths))
    else:
        model_path = MODEL_BASE_DIR / get_model_path(MODEL_CONFIG)

    # Get the QR-code paths.
    dataset_path = os.path.join(dataset_path, "scans")
    logging.info('Dataset path: %s', dataset_path)
    # logging.info(glob.glob(os.path.join(dataset_path, "*"))) # Debug
    logging.info('Getting QR-code paths...')
    qrcode_paths = glob.glob(os.path.join(dataset_path, "*"))
    logging.info('qrcode_paths: %d', len(qrcode_paths))
    assert len(qrcode_paths) != 0

    if EVAL_CONFIG.DEBUG_RUN and len(qrcode_paths) > EVAL_CONFIG.DEBUG_NUMBER_OF_SCAN:
        qrcode_paths = qrcode_paths[:EVAL_CONFIG.DEBUG_NUMBER_OF_SCAN]
        logging.info("Executing on %d qrcodes for FAST RUN", EVAL_CONFIG.DEBUG_NUMBER_OF_SCAN)

    logging.info('Paths for evaluation: \n\t' + '\n\t'.join(qrcode_paths))
    logging.info(len(qrcode_paths))

    # Is this a multiartifact model?
    if getattr(DATA_CONFIG, "N_ARTIFACTS", 1) > 1:
        samples_paths = create_multiartifact_paths_for_qrcodes(qrcode_paths, DATA_CONFIG)
        predictions = get_prediction_multiartifact(model_path, samples_paths)

        df = pd.DataFrame(predictions, columns=['artifacts', 'predicted', 'GT'])
        df['scantype'] = df.apply(extract_scantype, axis=1)
        df['qrcode'] = df.apply(extract_qrcode, axis=1)
        MAE = df.groupby(['qrcode', 'scantype']).mean()
        MAE['error'] = MAE.apply(avgerror, axis=1)

    else:  # Single-artifact

        # Get the pointclouds.
        logging.info("Getting Depthmap paths...")
        paths_evaluation = get_depthmap_files(qrcode_paths)
        del qrcode_paths

        logging.info("Using %d artifact files for evaluation.", len(paths_evaluation))

        new_paths_evaluation = paths_evaluation

        if FILTER_CONFIG is not None and FILTER_CONFIG.IS_ENABLED:
            standing = load_model(FILTER_CONFIG.NAME)
            new_paths_evaluation = filter_dataset_according_to_standing_lying(paths_evaluation, standing)

        logging.info("Creating dataset for training.")
        paths = new_paths_evaluation
        dataset = tf.data.Dataset.from_tensor_slices(paths)
        dataset_norm = dataset.map(lambda path: tf_load_pickle(path, DATA_CONFIG.NORMALIZATION_VALUE))

        # filter goodbad==delete
        if GOODBAD_IDX in DATA_CONFIG.TARGET_INDEXES:
            goodbad_index = DATA_CONFIG.TARGET_INDEXES.index(GOODBAD_IDX)
            dataset_norm = dataset_norm.filter(
                lambda _path, _depthmap, targets: targets[goodbad_index] != GOODBAD_DICT['delete'])

        dataset_norm = dataset_norm.cache()
        dataset_norm = dataset_norm.prefetch(tf.data.experimental.AUTOTUNE)
        temp_dataset_evaluation = dataset_norm
        del dataset_norm
        logging.info("Created dataset for training.")

        # Update new_paths_evaluation after filtering
        dataset_paths = temp_dataset_evaluation.map(lambda path, _depthmap, _targets: path)
        list_paths = list(dataset_paths.as_numpy_iterator())
        new_paths_evaluation = [x.decode() for x in list_paths]

        dataset_evaluation = temp_dataset_evaluation.map(lambda _path, depthmap, targets: (depthmap, targets))
        del temp_dataset_evaluation

        if RUN_IDS is not None:
            prediction_list_one = get_predictions_from_multiple_models(model_paths, dataset_evaluation)
        if RUN_ID is not None:
            prediction_list_one = get_prediction(model_path, dataset_evaluation)
        logging.info("Prediction made by model on the depthmaps...")
        logging.info(prediction_list_one)

        qrcode_list, scantype_list, artifact_list, prediction_list, target_list = get_column_list(
            new_paths_evaluation, prediction_list_one, DATA_CONFIG, FILTER_CONFIG)

        df = pd.DataFrame({
            'qrcode': qrcode_list,
            'artifact': artifact_list,
            'scantype': scantype_list,
            'GT': target_list if target_list[0].shape == tuple() else [el[0] for el in target_list],
            'predicted': prediction_list
        }, columns=RESULT_CONFIG.COLUMNS)
        logging.info("df.shape: %s", df.shape)

    df['GT'] = df['GT'].astype('float64')
    df['predicted'] = df['predicted'].astype('float64')

    if 'AGE_BUCKETS' in RESULT_CONFIG.keys():
        idx = DATA_CONFIG.TARGET_INDEXES.index(AGE_IDX)
        df[COLUMN_NAME_AGE] = [el[idx] for el in target_list]
    if SEX_IDX in DATA_CONFIG.TARGET_INDEXES:
        idx = DATA_CONFIG.TARGET_INDEXES.index(SEX_IDX)
        df[COLUMN_NAME_SEX] = [el[idx] for el in target_list]
    if GOODBAD_IDX in DATA_CONFIG.TARGET_INDEXES:
        idx = DATA_CONFIG.TARGET_INDEXES.index(GOODBAD_IDX)
        df[COLUMN_NAME_GOODBAD] = [el[idx] for el in target_list]

    df_grouped = df.groupby(['qrcode', 'scantype']).mean()
    logging.info("Mean Avg Error: %s", df_grouped)

    df_grouped['error'] = df_grouped.apply(avgerror, axis=1)

    descriptor = RUN_ID if RUN_ID else MODEL_CONFIG.EXPERIMENT_NAME

    csv_fpath = f"{OUTPUT_CSV_PATH}/{descriptor}.csv"
    logging.info("Calculate and save the results to %s", csv_fpath)
    calculate_and_save_results(df_grouped, EVAL_CONFIG.NAME, csv_fpath,
                               DATA_CONFIG, RESULT_CONFIG, fct=calculate_performance)

    sample_csv_fpath = f"{OUTPUT_CSV_PATH}/inaccurate_scans_{descriptor}.csv"
    df_grouped.to_csv(sample_csv_fpath, index=True)

    if 'AGE_BUCKETS' in RESULT_CONFIG.keys():
        csv_fpath = f"{OUTPUT_CSV_PATH}/age_evaluation_{descriptor}.csv"
        logging.info("Calculate and save age results to %s", csv_fpath)
        calculate_and_save_results(df_grouped, EVAL_CONFIG.NAME, csv_fpath,
                                   DATA_CONFIG, RESULT_CONFIG, fct=calculate_performance_age)
        png_fpath = f"{OUTPUT_CSV_PATH}/age_evaluation_scatter_{descriptor}.png"
        logging.info("Calculate and save scatterplot results to %s", png_fpath)
        draw_age_scatterplot(df, png_fpath)

    if (HEIGHT_IDX in DATA_CONFIG.TARGET_INDEXES
            and AGE_IDX in DATA_CONFIG.TARGET_INDEXES
            and descriptor != MODEL_CONFIG.EXPERIMENT_NAME):
        png_fpath = f"{OUTPUT_CSV_PATH}/stunting_diagnosis_{descriptor}.png"
        logging.info("Calculate zscores and save confusion matrix results to %s", png_fpath)
        start = time.time()
        draw_stunting_diagnosis(df, png_fpath)
        end = time.time()
        logging.info("Total time for Calculate zscores and save confusion matrix: %.2f", end - start)

    if (WEIGHT_IDX in DATA_CONFIG.TARGET_INDEXES
            and AGE_IDX in DATA_CONFIG.TARGET_INDEXES
            and descriptor != MODEL_CONFIG.EXPERIMENT_NAME):
        png_fpath = f"{OUTPUT_CSV_PATH}/wasting_diagnosis_{descriptor}.png"
        logging.info("Calculate and save wasting confusion matrix results to %s", png_fpath)
        start = time.time()
        draw_wasting_diagnosis(df, png_fpath)
        end = time.time()
        logging.info("Total time for Calculate zscores and save wasting confusion matrix: %.2f", end - start)

    if SEX_IDX in DATA_CONFIG.TARGET_INDEXES:
        csv_fpath = f"{OUTPUT_CSV_PATH}/sex_evaluation_{descriptor}.csv"
        logging.info("Calculate and save sex results to %s", csv_fpath)
        calculate_and_save_results(df_grouped, EVAL_CONFIG.NAME, csv_fpath,
                                   DATA_CONFIG, RESULT_CONFIG, fct=calculate_performance_sex)
    if GOODBAD_IDX in DATA_CONFIG.TARGET_INDEXES:
        csv_fpath = f"{OUTPUT_CSV_PATH}/goodbad_evaluation_{descriptor}.csv"
        logging.info("Calculate performance on bad/good scans and save results to %s", csv_fpath)
        calculate_and_save_results(df_grouped, EVAL_CONFIG.NAME, csv_fpath,
                                   DATA_CONFIG, RESULT_CONFIG, fct=calculate_performance_goodbad)

    if RESULT_CONFIG.USE_UNCERTAINTY:
        assert GOODBAD_IDX in DATA_CONFIG.TARGET_INDEXES
        assert COLUMN_NAME_GOODBAD in df
        assert RUN_IDS
        assert not RUN_ID

        # Sample one artifact per scan (qrcode, scantype combination)
        df_sample = df.groupby(['qrcode', 'scantype']).apply(lambda x: x.sample(1))

        # Prepare uncertainty prediction on these artifacts
        dataset_sample = prepare_sample_dataset(df_sample, dataset_path)

        # Predict uncertainty
        uncertainties = get_prediction_uncertainty_deepensemble(model_paths, dataset_sample)

        assert len(df_sample) == len(uncertainties)
        df_sample['uncertainties'] = uncertainties

        png_fpath = f"{OUTPUT_CSV_PATH}/uncertainty_distribution.png"
        draw_uncertainty_goodbad_plot(df_sample, png_fpath)

        df_sample_100 = df_sample.iloc[df_sample.index.get_level_values('scantype') == '100']
        png_fpath = f"{OUTPUT_CSV_PATH}/uncertainty_code100_distribution.png"
        draw_uncertainty_goodbad_plot(df_sample_100, png_fpath)

        png_fpath = f"{OUTPUT_CSV_PATH}/uncertainty_scatter_distribution.png"
        draw_uncertainty_scatterplot(df_sample, png_fpath)

        # Filter for scans with high certainty and calculate their accuracy/results
        df_sample['error'] = df_sample.apply(avgerror, axis=1).abs()
        df_sample_better_threshold = df_sample[df_sample['uncertainties'] < RESULT_CONFIG.UNCERTAINTY_THRESHOLD_IN_CM]
        csv_fpath = f"{OUTPUT_CSV_PATH}/uncertainty_smaller_than_{RESULT_CONFIG.UNCERTAINTY_THRESHOLD_IN_CM}cm.csv"
        logging.info("Uncertainty: For more certain than %.2f cm, calculate and save the results to %s",
                     RESULT_CONFIG.UNCERTAINTY_THRESHOLD_IN_CM, csv_fpath)
        calculate_and_save_results(df_sample_better_threshold, EVAL_CONFIG.NAME, csv_fpath,
                                   DATA_CONFIG, RESULT_CONFIG, fct=calculate_performance)

    # Done.
    run.complete()



import logging
import logging.config
from typing import List


logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d')

IDENTITY_MATRIX_4D = [1., 0., 0., 0.,
                      0., 1., 0., 0.,
                      0., 0., 1., 0.,
                      0., 0., 0., 1.]


def cross(a: list, b: list) -> list:
    """Cross product of two vectors"""
    c = [a[1] * b[2] - a[2] * b[1],
         a[2] * b[0] - a[0] * b[2],
         a[0] * b[1] - a[1] * b[0]]
    return c


def diff(a: list, b: list) -> list:
    """Difference of two vectors"""
    return [a[0] - b[0], a[1] - b[1], a[2] - b[2]]


def norm(v: list) -> list:
    """Vector normalize"""
    length = abs(v[0]) + abs(v[1]) + abs(v[2])
    if length == 0:
        length = 1
    return [v[0] / length, v[1] / length, v[2] / length]


def matrix_calculate(position: list, rotation: list) -> list:
    """Calculate a matrix image->world from device position and rotation"""

    output = IDENTITY_MATRIX_4D

    sqw = rotation[3] * rotation[3]
    sqx = rotation[0] * rotation[0]
    sqy = rotation[1] * rotation[1]
    sqz = rotation[2] * rotation[2]

    invs = 1 / (sqx + sqy + sqz + sqw)
    output[0] = (sqx - sqy - sqz + sqw) * invs
    output[5] = (-sqx + sqy - sqz + sqw) * invs
    output[10] = (-sqx - sqy + sqz + sqw) * invs

    tmp1 = rotation[0] * rotation[1]
    tmp2 = rotation[2] * rotation[3]
    output[1] = 2.0 * (tmp1 + tmp2) * invs
    output[4] = 2.0 * (tmp1 - tmp2) * invs

    tmp1 = rotation[0] * rotation[2]
    tmp2 = rotation[1] * rotation[3]
    output[2] = 2.0 * (tmp1 - tmp2) * invs
    output[8] = 2.0 * (tmp1 + tmp2) * invs

    tmp1 = rotation[1] * rotation[2]
    tmp2 = rotation[0] * rotation[3]
    output[6] = 2.0 * (tmp1 + tmp2) * invs
    output[9] = 2.0 * (tmp1 - tmp2) * invs

    output[12] = -position[0]
    output[13] = -position[1]
    output[14] = -position[2]
    return output


def matrix_transform_point(point: list, matrix: list) -> list:
    """Transformation of point by matrix"""
    output = [0, 0, 0, 1]
    output[0] = point[0] * matrix[0] + point[1] * matrix[4] + point[2] * matrix[8] + matrix[12]
    output[1] = point[0] * matrix[1] + point[1] * matrix[5] + point[2] * matrix[9] + matrix[13]
    output[2] = point[0] * matrix[2] + point[1] * matrix[6] + point[2] * matrix[10] + matrix[14]
    output[3] = point[0] * matrix[3] + point[1] * matrix[7] + point[2] * matrix[11] + matrix[15]

    output[0] /= abs(output[3])
    output[1] /= abs(output[3])
    output[2] /= abs(output[3])
    output[3] = 1
    return output


def parse_numbers(line: str) -> List[float]:
    """Parse line of numbers

    Args:
        line: Example: "0.6786797 0.90489584 0.49585155 0.5035042"

    Return:
        numbers: [0.6786797, 0.90489584, 0.49585155, 0.5035042]
    """
    return [float(value) for value in line.split(' ')]



import matplotlib.pyplot as plt
import pandas as pd
import os
from decimal import Decimal as D
import json
import numpy as np

module_dir = os.path.split(os.path.abspath(__file__))[0]


class Observation(object):
    def __init__(self, chart, weight, muac,
                 age_in_days, sex, height):

        self.chart = chart
        self.weight = weight
        self.muac = muac
        self.age_in_days = age_in_days
        self.sex = sex
        self.height = height

        self.table_chart = None
        self.table_age = None
        self.table_sex = None
        if self.chart in ['wfl', 'wfh']:
            if self.height in ['', ' ', None]:
                raise Exception('no length or height')
        if self.chart in ['wfa']:
            if self.weight in ['', ' ', None]:
                raise Exception('no weight')

    def get_values(self, growth):
        table_name = self.resolve_table()
        table = getattr(growth, table_name)
        if self.chart in ["wfh", "wfl"]:
            if D(self.height) < 45:
                raise Exception("too short")
            if D(self.height) > 120:
                raise TypeError("too tall")
            closest_height = float("{0:.1f}". format(D(self.height)))
            scores = table.get(str(closest_height))
            if scores is not None:
                return scores
            raise TypeError(
                "SCORES NOT FOUND FOR HEIGHT :%s", (self.closest_height))

        elif self.chart in ['wfa', 'lhfa']:
            scores = table.get(str(self.age_in_days))
            if scores is not None:
                return scores
            raise TypeError(
                "SCORES NOT FOUND BY DAY : %s", self.age_in_days)

    def resolve_table(self):

        if self.chart == 'wfl' and D(self.height) > 110:
            self.table_chart = 'wfh'
            self.table_age = '2_5'
        elif self.chart == 'wfh' and D(self.height) < 65:
            self.table_chart = 'wfl'
            self.table_age = '0_2'
        else:
            self.table_chart = self.chart
            if self.chart == 'wfl':
                self.table_age = '0_2'
            if self.table_chart == 'wfh':
                self.table_age = '2_5'

        if self.sex == 'M':
            self.table_sex = 'boys'
        if self.sex == 'F':
            self.table_sex = 'girls'

        if self.chart in ["wfa", "lhfa"]:
            self.table_age = "0_5"
            self.table_chart = self.chart

        table = "%(table_chart)s_%(table_sex)s_%(table_age)s" %\
                {"table_chart": self.table_chart,
                 "table_sex": self.table_sex,
                 "table_age": self.table_age}
        return table


class Calculator(object):

    def __reformat_table(self, table_name):
        list_of_dicts = getattr(self, table_name)
        if 'Length' in list_of_dicts[0]:
            field_name = 'Length'
        elif 'Height' in list_of_dicts[0]:
            field_name = 'Height'
        elif 'Day' in list_of_dicts[0]:
            field_name = 'Day'
        else:
            raise Exception('error loading: %s' % table_name)
        new_dict = {'field_name': field_name}
        for d in list_of_dicts:
            new_dict.update({d[field_name]: d})
        setattr(self, table_name, new_dict)

    def __init__(self):

        WHO_tables = [
            'wfl_boys_0_2_zscores.json', 'wfl_girls_0_2_zscores.json',
            'wfh_boys_2_5_zscores.json', 'wfh_girls_2_5_zscores.json',
            'wfa_boys_0_5_zscores.json', 'wfa_girls_0_5_zscores.json',
            'lhfa_boys_0_5_zscores.json', 'lhfa_girls_0_5_zscores.json',
        ]

        table_dir = os.path.join(module_dir, 'tables')
        table_to_load = WHO_tables

        for table in table_to_load:
            table_file = os.path.join(table_dir, table)
            with open(table_file, 'r') as f:
                table_name, underscore, zscore_part = table.split('.')[0].rpartition('_')
                setattr(self, table_name, json.load(f))
                self.__reformat_table(table_name)

    def zScore_wfa(self, weight=None, muac=None, age_in_days=None, sex=None, height=None):
        return self.z_score_measurement('wfa', weight=weight, muac=None, age_in_days=age_in_days, sex=sex, height=None)

    def zScore_wfl(self, weight=None, muac=None, age_in_days=None, sex=None, height=None):
        if D(age_in_days) > 731:
            return self.zScore_wfh(weight, muac, age_in_days, sex, height)
        return self.z_score_measurement('wfl', weight=weight, muac=None,
                                        age_in_days=age_in_days, sex=sex, height=height)

    def zScore_wfh(self, weight=None, muac=None, age_in_days=None, sex=None, height=None):
        if D(age_in_days) <= 731:
            return self.zScore_wfl(weight, muac, age_in_days, sex, height)
        return self.z_score_measurement('wfh', weight=weight, muac=None,
                                        age_in_days=age_in_days, sex=sex, height=height)

    def zScore_lhfa(self, weight=None, muac=None, age_in_days=None, sex=None, height=None):
        return self.z_score_measurement('lhfa', weight=None, muac=None, age_in_days=age_in_days, sex=sex, height=height)

    def zScore(self, weight=None, muac=None, age_in_days=None, sex=None, height=None):
        wfa = self.zScore_wfa(
            weight=weight, age_in_days=age_in_days, sex=sex)
        if D(age_in_days) > 731:
            dummy = 'Z_score_WFH'
            wfl = self.zScore_wfh(weight=weight, age_in_days=age_in_days, sex=sex, height=height)
        else:
            dummy = 'Z_score_WFL'
            wfl = self.zScore_wfl(weight=weight, age_in_days=age_in_days, sex=sex, height=height)
        lhfa = self.zScore_lhfa(age_in_days=age_in_days, sex=sex, height=height)
        zscore = json.dumps({'Z_score_WFA': wfa, dummy: wfl, "Z_score_HFA": lhfa})
        return zscore

    def zScore_withclass(self, weight=None, muac=None, age_in_days=None, sex=None, height=None):
        wfa = self.zScore_wfa(
            weight=weight, age_in_days=age_in_days, sex=sex)
        if wfa < -3:
            class_wfa = 'Severly Under-weight'
        elif wfa >= -3 and wfa < -2:
            class_wfa = 'Moderately Under-weight'
        else:
            class_wfa = 'Healthy'

        if D(age_in_days) > 24:
            dummy = 'Z_score_WFH'
            dummy_class = "Class_WFH"
            wfl = self.zScore_wfh(
                weight=weight, age_in_days=age_in_days, sex=sex, height=height)
        else:
            dummy = 'Z_score_WFL'
            dummy_class = "Class_WFL"
            wfl = self.zScore_wfl(
                weight=weight, age_in_days=age_in_days, sex=sex, height=height)
        class_wfl = self.SAM_MAM(weight, muac, age_in_days, sex, height)

        lhfa = self.zScore_lhfa(
            age_in_days=age_in_days, sex=sex, height=height)
        if lhfa < -3:
            class_lhfa = 'Severly Stunted'
        elif lhfa >= -3 and lhfa < -2:
            class_lhfa = 'Moderately Stunted'
        else:
            class_lhfa = 'Healthy'

        zscore = json.dumps({'Z_score_WFA': wfa, 'Class_WFA': class_wfa, dummy: wfl,
                             dummy_class: class_wfl, 'Z_score_HFA': lhfa, 'Class_HFA': class_lhfa})
        return zscore

    def SAM_MAM(self, weight=None, muac=None, age_in_days=None, sex=None, height=None):
        assert muac is not None

        if D(age_in_days) > 731:
            wfl = self.zScore_wfh(
                weight=weight, age_in_days=age_in_days, sex=sex, height=height)
        else:
            wfl = self.zScore_wfl(
                weight=weight, age_in_days=age_in_days, sex=sex, height=height)
        if wfl < -3 or D(muac) < 11.5:
            return "SAM"
        elif (wfl >= -3 and wfl < -2) or D(muac) < 12.5:
            return "MAM"
        else:
            return "Healthy"

    def z_score_measurement(self, chart, weight, muac, age_in_days, sex, height):
        assert sex is not None
        assert sex.upper() in ["M", "F"]
        assert age_in_days is not None
        assert chart is not None
        assert chart.lower() in ['wfa', 'wfh', 'wfl', 'lhfa']

        obs = Observation(chart, weight, muac, age_in_days,
                          sex, height)
        value = obs.get_values(self)

        skew = D(value.get("L"))
        median = D(value.get("M"))
        cofficient = D(value.get("S"))

        ###
        #  Z score
        #          [y/M(t)]^L(t) - 1
        #   Zind =  -----------------
        #               S(t)L(t)
        ###
        if chart == 'wfa' or chart == 'wfl' or chart == 'wfh':
            y = D(weight)
        elif chart == 'lhfa':
            y = D(height)

        numerator = (y / median)**skew - D(1.0)
        denomenator = skew * cofficient
        zScore = numerator / denomenator

        #            _
        #           |
        #           |       Zind            if |Zind| <= 3
        #           |
        #           |
        #           |       y - SD3pos
        #   Zind* = | 3 + ( ----------- )   if Zind > 3
        #           |         SD23pos
        #           |
        #           |
        #           |
        #           |        y - SD3neg
        #           | -3 + ( ----------- )  if Zind < -3
        #           |          SD23neg
        #           |
        #           |_

        def calc_stdev(sd):
            value = (1 + (skew * cofficient * sd))**(1 / skew)
            stdev = median * value
            return stdev

        if D(zScore) > D(3):
            SD2pos = calc_stdev(2)
            SD3pos = calc_stdev(3)

            SD23pos = SD3pos - SD2pos

            zScore = 3 + ((y - SD3pos) / SD23pos)

            zScore = float(zScore.quantize(D('0.01')))

        elif D(zScore) < -3:
            SD2neg = calc_stdev(-2)
            SD3neg = calc_stdev(-3)

            SD23neg = SD2neg - SD3neg

            zScore = -3 + ((y - SD3neg) / SD23neg)
            zScore = float(zScore.quantize(D('0.01')))

        else:
            zScore = float(zScore.quantize(D('0.01')))

        return zScore


class Chart():

    def __init__(self):
        super().__init__()

    def zScore_wfa_chart(self, weight=None, muac=None, age_in_days=None, sex=None, height=None):
        if sex == 'M':
            return self.zScore_chart('wfa_boys', weight, age_in_days, sex)
        if sex == 'F':
            return self.zScore_chart('wfa_girls', weight, age_in_days, sex)

    def zScore_lhfa_chart(self, weight=None, muac=None, age_in_days=None, sex=None, height=None):
        if sex == 'M':
            return self.zScore_chart('lhfa_boys', weight, age_in_days, sex, height)
        if sex == 'F':
            return self.zScore_chart('lhfa_girls', weight, age_in_days, sex, height)

    def zScore_wfl_chart(self, weight=None, muac=None, age_in_days=None, sex=None, height=None):
        if max(age_in_days) <= 731:
            if sex == 'M':
                return self.zScore_chart('wfl_boys', weight, age_in_days, sex, height)
            if sex == 'F':
                return self.zScore_chart('wfl_girls', weight, age_in_days, sex, height)
        else:
            raise Exception("MAX AGE IS MORE THEN 2 YEARS. Use zScore_wfh_full_chart")

    def zScore_wfh_chart(self, weight=None, muac=None, age_in_days=None, sex=None, height=None):
        if min(age_in_days) >= 731:
            if sex == 'M':
                return self.zScore_chart('wfh_boys_2_5', weight, age_in_days, sex, height)
            if sex == 'F':
                return self.zScore_chart('wfh_girls_2_5', weight, age_in_days, sex, height)
        else:
            raise Exception("MIN AGE IS LESS THEN 2 YEARS. USE zScore_wfh_full_chart")

    def zScore_wfh_full_chart(self, weight=None, muac=None, age_in_days=None, sex=None, height=None):
        if sex == 'M':
            return self.zScore_chart('wfh_boys_0_5', weight, age_in_days, sex, height)
        if sex == 'F':
            return self.zScore_chart('wfh_girls_0_5', weight, age_in_days, sex, height)

    def zScore_chart(self, chart=None, weight=None, age_in_days=None, sex=None, height=None):
        if chart == 'wfa_boys' or chart == 'wfa_girls' or chart == 'lhfa_boys' or chart == 'lhfa_girls':
            df = pd.read_json(str(module_dir) + '/tables/' + chart + '_0_5_zscores.json')
            x = df['Day'].values.tolist()
        elif chart == 'wfl_boys' or chart == 'wfl_girls':
            df = pd.read_json(str(module_dir) + 'tables/' + chart + '_0_2_zscores.json')
            x = df['Length'].values.tolist()
        elif chart == 'wfh_boys_2_5' or chart == 'wfh_girls_2_5':
            df = pd.read_json(str(module_dir) + 'tables/' + chart + '_zscores.json')
            x = df['Height'].values.tolist()
        elif chart == 'wfh_boys_0_5' or chart == 'wfh_girls_0_5':
            df = pd.read_json(str(module_dir) + 'tables/' + chart + '_zscores.json')
            x = df['Height'].values.tolist()

        a = df['SD3'].values.tolist()
        b = df['SD2'].values.tolist()
        c = df['SD1'].values.tolist()
        d = df['SD0'].values.tolist()
        e = df['SD1neg'].values.tolist()
        f = df['SD2neg'].values.tolist()
        g = df['SD3neg'].values.tolist()

        plt.figure(figsize=(15, 7))

        plt.plot(x, a, '-r', color='black', linewidth=0.8, label='3')
        plt.plot(x, b, color='red', linewidth=0.8, label='-2')
        plt.plot(x, c, color=[0.9100, 0.4100, 0.1700],
                 linewidth=0.8, label='-1')
        plt.plot(x, d, color='green', linewidth=0.8, label='0')
        plt.plot(x, e, color=[0.9100, 0.4100, 0.1700],
                 linewidth=0.8, label='-1')
        plt.plot(x, f, color='red', linewidth=0.8, label='-2')
        plt.plot(x, g, color='black', linewidth=0.8, label='-3')

        if chart == 'wfa_boys' or chart == 'wfa_girls':
            plt.plot(age_in_days, weight, 'o-', label='z score')
            for i in range(len(age_in_days)):
                plt.text(age_in_days[i], weight[i], str(Calculator().zScore_wfa(
                    weight=str(weight[i]), age_in_days=str(age_in_days[i]), sex=sex)),
                    horizontalalignment='center', color='black')
            plt.ylabel('Weight(km)')
            plt.xlabel('Age in days')
            plt.xticks(np.arange(0, 1865, 90))
            plt.yticks(np.arange(0, 30, 2))
            plt.title('Z score weight vs age')
        elif chart == 'lhfa_boys' or chart == 'lhfa_girls':
            plt.plot(age_in_days, height, 'o-', label='z score')
            for i in range(len(age_in_days)):
                plt.text(age_in_days[i], height[i], str(Calculator().zScore_lhfa(
                    height=str(height[i]), age_in_days=str(age_in_days[i]), sex=sex)),
                    horizontalalignment='center', color='black')
            plt.ylabel('Height(cm)')
            plt.xlabel('Age in days')
            plt.xticks(np.arange(0, 1865, 90))
            plt.yticks(np.arange(35, 110, 5))
            plt.title('Z score height vs age')
        elif chart == 'wfl_boys' or chart == 'wfl_girls':
            plt.plot(height, weight, 'o-', label='z score')
            for i in range(len(height)):
                plt.text(height[i], weight[i], str(Calculator().zScore_wfl(
                    height=str(height[i]), age_in_days=str(age_in_days[i]), sex=sex, weight=str(weight[i]))),
                    horizontalalignment='center', color='black')
            plt.ylabel('Weight(kg)')
            plt.xlabel('Height(cm)')
            plt.xticks(np.arange(45, 110, 5))
            plt.yticks(np.arange(0, 24, 2))
            plt.title('Z score weight vs length')
        elif chart == 'wfh_boys_2_5' or chart == 'wfh_girls_2_5':
            plt.plot(height, weight, 'o-', label='z score')
            for i in range(len(height)):
                plt.text(height[i], weight[i], str(Calculator().zScore_wfh(
                    height=str(height[i]), age_in_days=str(age_in_days[i]), sex=sex, weight=str(weight[i]))),
                    horizontalalignment='center', color='black')
            plt.ylabel('Weight(kg)')
            plt.xlabel('Height(cm)')
            plt.xticks(np.arange(65, 120, 5))
            plt.yticks(np.arange(0, 32, 2))
            plt.title('Z score weight vs length')

        elif chart == 'wfh_boys_0_5' or chart == 'wfh_girls_0_5':
            plt.plot(height, weight, 'o-', label='z score')
            for i in range(len(height)):
                plt.text(height[i], weight[i], str(Calculator().zScore_wfh(
                    height=str(height[i]), age_in_days=str(age_in_days[i]), sex=sex, weight=str(weight[i]))),
                    horizontalalignment='center', color='black')
            plt.ylabel('Weight(kg)')
            plt.xlabel('Height(cm)')
            plt.xticks(np.arange(45, 120, 5))
            plt.yticks(np.arange(0, 32, 2))
            plt.title('Z score weight vs length')

        plt.grid(color='r', linestyle='dashed', linewidth=0.5)
        plt.legend()

        return plt



import os
import shutil
import sys
from os import walk
import logging
import logging.config
from pathlib import Path
import functools

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Button

from depthmap import Depthmap
from exporter import export_obj, export_pcd
from visualisation import render_plot

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d')

# click on data
LAST = [0, 0, 0]
# INDEX of the current depthmap/rgb frame
INDEX = 0
# current depthmap
DMAP = 0


def onclick(event):
    global DMAP
    global LAST
    if event.xdata is not None and event.ydata is not None:
        x = int(event.ydata)
        y = DMAP.height - int(event.xdata) - 1
        if x > 1 and y > 1 and x < DMAP.width - 2 and y < DMAP.height - 2:
            depth = DMAP.parse_depth(x, y)
            if depth:
                res = DMAP.convert_2d_to_3d(1, x, y, depth)
                if res:
                    diff = [LAST[0] - res[0], LAST[1] - res[1], LAST[2] - res[2]]
                    dst = np.sqrt(diff[0] * diff[0] + diff[1] * diff[1] + diff[2] * diff[2])
                    res.append(dst)
                    logging.info('x=%s, y=%s, depth=%s, diff=%s', str(res[0]), str(res[1]), str(res[2]), str(res[3]))
                    LAST[0] = res[0]
                    LAST[1] = res[1]
                    LAST[2] = res[2]
                    return
            logging.info('no valid data')


def export_object(event):
    global DMAP
    fname = f'output{INDEX}.obj'
    export_obj('export/' + fname, DMAP, triangulate=True)


def export_pointcloud(event):
    global DMAP
    fname = f'output{INDEX}.pcd'
    export_pcd('export/' + fname, DMAP)


def next_click(event, calibration_file: str, depthmap_dir: str):
    global INDEX
    INDEX = INDEX + 1
    if (INDEX == size):
        INDEX = 0
    show(depthmap_dir, calibration_file)


def prev_click(event, calibration_file: str, depthmap_dir: str):
    global INDEX
    INDEX = INDEX - 1
    if (INDEX == -1):
        INDEX = size - 1
    show(depthmap_dir, calibration_file)


def show(depthmap_dir: str, calibration_file: str):
    global DMAP
    fig.canvas.manager.set_window_title(depth_filenames[INDEX])
    rgb_filename = rgb_filenames[INDEX] if rgb_filenames else 0
    DMAP = Depthmap.create_from_file(depthmap_dir, depth_filenames[INDEX], rgb_filename, calibration_file)

    angle = DMAP.get_angle_between_camera_and_floor()
    logging.info('angle between camera and floor is %f', angle)

    render_plot(DMAP)
    plt.show()


if __name__ == "__main__":
    # Prepare
    if len(sys.argv) != 3:
        logging.info('You did not enter depthmap_dir folder and calibration file path')
        logging.info('E.g.: python toolkit.py depthmap_dir calibration_file')
        sys.exit(1)

    depthmap_dir = sys.argv[1]
    calibration_file = sys.argv[2]

    depth_filenames = []
    for (dirpath, dirnames, filenames) in walk(Path(depthmap_dir) / 'depth'):
        depth_filenames.extend(filenames)
    depth_filenames.sort()

    rgb_filenames = []
    for (dirpath, dirnames, filenames) in walk(Path(depthmap_dir) / 'rgb'):
        rgb_filenames.extend(filenames)
    rgb_filenames.sort()

    # Clear export folder
    try:
        shutil.rmtree('export')
    except BaseException:
        print('no previous data to delete')
    os.mkdir('export')

    # Show viewer
    INDEX = 0
    size = len(depth_filenames)
    fig = plt.figure()
    fig.canvas.mpl_connect('button_press_event', functools.partial(onclick))
    bprev = Button(plt.axes([0.0, 0.0, 0.1, 0.075]), '<<', color='gray')
    bprev.on_clicked(functools.partial(prev_click, calibration_file=calibration_file, depthmap_dir=depthmap_dir))
    bnext = Button(plt.axes([0.9, 0.0, 0.1, 0.075]), '>>', color='gray')
    bnext.on_clicked(functools.partial(next_click, calibration_file=calibration_file, depthmap_dir=depthmap_dir))
    bexport_obj = Button(plt.axes([0.3, 0.0, 0.2, 0.05]), 'Export OBJ', color='gray')
    bexport_obj.on_clicked(functools.partial(export_object))
    bexport_pcd = Button(plt.axes([0.5, 0.0, 0.2, 0.05]), 'Export PCD', color='gray')
    bexport_pcd.on_clicked(functools.partial(export_pointcloud))
    background = Button(plt.axes([0.0, 0.0, 1.0, 1.0]), '', color='white')
    show(depthmap_dir, calibration_file)



import os

from bunch import Bunch

CONFIG_NAME = os.path.splitext(os.path.basename(__file__))[0]

# Details of model used for evaluation
MODEL_CONFIG = Bunch(dict(
    EXPERIMENT_NAME='q1-ensemble-warmup',
    RUN_IDS=[
        'q1-ensemble-warmup_1610544610_eb44bfe2', 'q1-ensemble-warmup_1610547587_7ca932c3',
        # 'q1-ensemble-warmup_1610547669_5b789bd1', 'q1-ensemble-warmup_1610547705_f2141d0f',
        # 'q1-ensemble-warmup_1610547744_d2b42ce5', 'q1-ensemble-warmup_1610547780_2f000a25',
        # 'q1-ensemble-warmup_1610547816_c3f815df', 'q1-ensemble-warmup_1610547892_8ee6ff49',
        # 'q1-ensemble-warmup_1610547928_b9519b6a', 'q1-ensemble-warmup_1610547986_ad0186b8',
        # 'q1-ensemble-warmup_1610548023_99ac6060', 'q1-ensemble-warmup_1610548064_afefd4e4',
        # 'q1-ensemble-warmup_1610548106_69993d24', 'q1-ensemble-warmup_1610548137_a8c52d63',
        # 'q1-ensemble-warmup_1610548168_914ce1f6', 'q1-ensemble-warmup_1610548209_9692a253',
    ],
    INPUT_LOCATION='outputs',
    NAME='best_model.ckpt',
))


EVAL_CONFIG = Bunch(dict(
    # Name of evaluation
    NAME='q3-depthmap-plaincnn-height-95k-run_03',

    # Experiment in Azure ML which will be used for evaluation
    EXPERIMENT_NAME="QA-pipeline",
    CLUSTER_NAME="gpu-cluster",

    # Used for Debug the QA pipeline
    DEBUG_RUN=False,

    # Will run eval on specified # of scan instead of full dataset
    DEBUG_NUMBER_OF_SCAN=5,

    SPLIT_SEED=0,
))

# Details of Evaluation Dataset
DATA_CONFIG = Bunch(dict(
    NAME='anon-realtime-testdata',  # Name of evaluation dataset

    IMAGE_TARGET_HEIGHT=240,
    IMAGE_TARGET_WIDTH=180,

    BATCH_SIZE=512,  # Batch size for evaluation
    NORMALIZATION_VALUE=7.5,

    TARGET_INDEXES=[0, 5],  # 0 is height, 1 is weight.
    CODES=['100', '101', '102', '200', '201', '202']
))


# Result configuration for result generation after evaluation is done
RESULT_CONFIG = Bunch(dict(
    # Error margin on various ranges
    #EVALUATION_ACCURACIES = [.2, .4, .8, 1.2, 2., 2.5, 3., 4., 5., 6.]
    ACCURACIES=[.2, .4, .6, 1, 1.2, 2., 2.5, 3., 4., 5., 6.],  # 0.2cm, 0.4cm, 0.6cm, 1cm, ...
    ACCURACY_MAIN_THRESH=1.0,
    COLUMNS=['qrcode', 'artifact', 'scantype', 'GT', 'predicted'],

    USE_UNCERTAINTY=True,  # Flag to enable model uncertainty calculation
    UNCERTAINTY_THRESHOLD_IN_CM=4.,

    TARGET_INDEXES=[0],
    # path of csv file in the experiment which final result is stored
    SAVE_PATH=f'./outputs/{CONFIG_NAME}',
))



import zipfile
import logging
import logging.config
import math

from pathlib import Path
import statistics
import numpy as np
from PIL import Image
from typing import List, Tuple

from depthmap_utils import (
    matrix_calculate, IDENTITY_MATRIX_4D, parse_numbers, diff, cross, norm, matrix_transform_point)
from constants import EXTRACTED_DEPTH_FILE_NAME, MASK_FLOOR, MASK_CHILD

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d')


TOOLKIT_DIR = Path(__file__).parents[0].absolute()


def extract_depthmap(depthmap_dir: str, depthmap_fname: str):
    """Extract depthmap from given file"""
    with zipfile.ZipFile(Path(depthmap_dir) / 'depth' / depthmap_fname, 'r') as zip_ref:
        zip_ref.extractall(TOOLKIT_DIR)
    return TOOLKIT_DIR / EXTRACTED_DEPTH_FILE_NAME


class Depthmap:
    """Depthmap and optional RGB

    Args:
        intrinsics (np.array): Camera intrinsics
        width (int): Width of the depthmap
        height (int): Height of the depthmap
        data (bytes): pixel_data
        depth_scale (float): Scalar to scale depthmap pixel to meters
        max_confidence (float): Confidence is amount of IR light reflected
                                (e.g. 0 to 255 in Lenovo, new standard is 0 to 7)
                                This is actually an int.
        matrix (list): Header contains a pose (= position and rotation)
                       - matrix is a list represenation of this pose
                       - can be used to project into a different space
        rgb_fpath (str): Path to RGB file (e.g. to the jpg)
        rgb_array (np.array): RGB data
    """
    def __init__(
            self,
            intrinsics,
            width,
            height,
            data,
            depth_scale,
            max_confidence,
            matrix,
            rgb_fpath,
            rgb_array):
        self.intrinsics = intrinsics
        self.width = width
        self.height = height
        self.data = data
        self.depth_scale = depth_scale
        self.max_confidence = max_confidence
        self.matrix = matrix
        self.rgb_fpath = rgb_fpath
        self.rgb_array = rgb_array

    @property
    def has_rgb(self) -> bool:
        """Bool that indicates if the object has RGB data"""
        return self.rgb_array is not None

    @classmethod
    def create_from_file(cls,
                         depthmap_dir: str,
                         depthmap_fname: str,
                         rgb_fname: str,
                         calibration_file: str):

        # read depthmap data
        path = extract_depthmap(depthmap_dir, depthmap_fname)
        with open(path, 'rb') as f:
            line = f.readline().decode().strip()
            header = line.split('_')
            res = header[0].split('x')
            width = int(res[0])
            height = int(res[1])
            depth_scale = float(header[1])
            max_confidence = float(header[2])
            if len(header) >= 10:
                position = (float(header[7]), float(header[8]), float(header[9]))
                rotation = (float(header[3]), float(header[4]), float(header[5]), float(header[6]))
                matrix = matrix_calculate(position, rotation)
            else:
                matrix = IDENTITY_MATRIX_4D
            data = f.read()
            f.close()

        # read rgb data
        if rgb_fname:
            rgb_fpath = depthmap_dir + '/rgb/' + rgb_fname
            pil_im = Image.open(rgb_fpath)
            pil_im = pil_im.resize((width, height), Image.ANTIALIAS)
            rgb_array = np.asarray(pil_im)
        else:
            rgb_fpath = rgb_fname
            rgb_array = None

        # read calibration file
        intrinsics = parse_calibration(calibration_file)

        return cls(intrinsics,
                   width,
                   height,
                   data,
                   depth_scale,
                   max_confidence,
                   matrix,
                   rgb_fpath,
                   rgb_array
                   )

    def calculate_normal_vector(self, x: float, y: float) -> list:
        """Calculate normal vector of depthmap point based on neightbors"""

        # Get depth of the neightbor pixels
        depth_center = self.parse_depth_smoothed(x, y)
        depth_x_minus = self.parse_depth_smoothed(x - 1, y)
        depth_y_minus = self.parse_depth_smoothed(x, y - 1)

        # Create a triangle from neighbor points
        point_a = self.convert_2d_to_3d_oriented(1, x, y, depth_center)
        point_b = self.convert_2d_to_3d_oriented(1, x - 1, y, depth_x_minus)
        point_c = self.convert_2d_to_3d_oriented(1, x, y - 1, depth_y_minus)

        # Calculate a normal of the triangle
        vector_u = diff(point_a, point_b)
        vector_v = diff(point_a, point_c)
        normal = cross(vector_u, vector_v)

        # Ensure the normal has a length of one
        return norm(normal)

    def convert_2d_to_3d(self, sensor: int, x: float, y: float, depth: float) -> list:
        """Convert point in pixels into point in meters"""
        fx = self.intrinsics[sensor][0] * float(self.width)
        fy = self.intrinsics[sensor][1] * float(self.height)
        cx = self.intrinsics[sensor][2] * float(self.width)
        cy = self.intrinsics[sensor][3] * float(self.height)
        tx = (x - cx) * depth / fx
        ty = (y - cy) * depth / fy
        return [tx, ty, depth]

    def convert_2d_to_3d_oriented(self, sensor: int, x: float, y: float, depth: float) -> list:
        """Convert point in pixels into point in meters (applying rotation)"""
        res = self.convert_2d_to_3d(sensor, x, y, depth)
        if not res:
            return res

        # special case for Google Tango devices with different rotation
        if self.width == 180 and self.height == 135:
            res = [res[0], -res[1], res[2]]
        else:
            res = [-res[0], res[1], res[2]]
        try:
            res = matrix_transform_point(res, self.matrix)
            res = [res[0], -res[1], res[2]]
        except NameError:
            pass
        return res

    def detect_child(self, floor: float) -> Tuple[np.array, float]:

        # mask the floor
        mask = np.zeros((self.width, self.height))
        for x in range(self.width):
            for y in range(self.height):
                depth = self.parse_depth(x, y)
                if not depth:
                    continue
                normal = self.calculate_normal_vector(x, y)
                point = self.convert_2d_to_3d_oriented(1, x, y, depth)
                if abs(normal[1]) > 0.5:
                    if abs(point[1] - floor) < 0.1:
                        mask[x][y] = MASK_FLOOR

        # highlight the focused child/object using seed algorithm
        highest = floor
        dirs = [[-1, 0], [1, 0], [0, -1], [0, 1]]
        pixel = [int(self.width / 2), int(self.height / 2)]
        stack = [pixel]
        while len(stack) > 0:

            # get a next pixel from the stack
            pixel = stack.pop()
            depth_center = self.parse_depth(pixel[0], pixel[1])

            # add neighbor points (if there is no floor and they are connected)
            if mask[pixel[0]][pixel[1]] == 0:
                for direction in dirs:
                    pixel_dir = [pixel[0] + direction[0], pixel[1] + direction[1]]
                    depth_dir = self.parse_depth(pixel_dir[0], pixel_dir[1])
                    if depth_dir > 0 and (depth_dir - depth_center) < 0.1:
                        stack.append(pixel_dir)

            # update the highest point
            point = self.convert_2d_to_3d_oriented(1, pixel[0], pixel[1], depth_center)
            highest = max(highest, point[1])

            # update the mask
            mask[pixel[0]][pixel[1]] = MASK_CHILD

        return mask, highest

    def get_angle_between_camera_and_floor(self) -> float:
        """Calculate an angle between camera and floor based on device pose"""
        centerx = float(self.width / 2)
        centery = float(self.height / 2)
        vector = self.convert_2d_to_3d_oriented(1, centerx, centery, 1.0)
        angle = 90 + math.degrees(math.atan2(vector[0], vector[1]))
        return angle

    def get_floor_level(self) -> float:
        """Calculate an altitude of the floor in the world coordinates"""
        altitudes = []
        for x in range(self.width):
            for y in range(self.height):
                normal = self.calculate_normal_vector(x, y)
                if abs(normal[1]) > 0.5:
                    depth = self.parse_depth(x, y)
                    point = self.convert_2d_to_3d_oriented(1, x, y, depth)
                    altitudes.append(point[1])
        return statistics.median(altitudes)

    def parse_confidence(self, tx: int, ty):
        """Get confidence of the point in scale 0-1"""
        return self.data[(int(ty) * self.width + int(tx)) * 3 + 2] / self.max_confidence

    def parse_depth(self, tx: int, ty: int) -> float:
        """Get depth of the point in meters"""
        if tx < 1 or ty < 1 or tx >= self.width or ty >= self.height:
            return 0.
        depth = self.data[(int(ty) * self.width + int(tx)) * 3 + 0] << 8
        depth += self.data[(int(ty) * self.width + int(tx)) * 3 + 1]
        depth *= self.depth_scale
        return depth

    def parse_depth_smoothed(self, tx: int, ty) -> float:
        """Get average depth value from neighboring pixels"""
        depth_center = self.parse_depth(tx, ty)
        depth_x_minus = self.parse_depth(tx - 1, ty)
        depth_x_plus = self.parse_depth(tx + 1, ty)
        depth_y_minus = self.parse_depth(tx, ty - 1)
        depth_y_plus = self.parse_depth(tx, ty + 1)
        depths = [depth_x_minus, depth_x_plus, depth_y_minus, depth_y_plus, depth_center]
        return sum(depths) / len(depths)

    def convert_3d_to_2d(self, sensor: int, x: float, y: float, depth: float) -> list:
        """Convert point in meters into point in pixels

        Args:
            sensor: Tells if this is ToF sensor or RGB sensor
            x: X-pos in m
            y: Y-pos in m
            depth: distance from sensor to object at (x, y)

        Returns:
            tx, ty, depth
        """
        return convert_3d_to_2d(self.intrinsics[sensor], x, y, depth, self.width, self.height)


def convert_3d_to_2d(intrinsics: list, x: float, y: float, depth: float, width: int, height: int):
    fx = intrinsics[0] * float(width)
    fy = intrinsics[1] * float(height)
    cx = intrinsics[2] * float(width)
    cy = intrinsics[3] * float(height)
    tx = x * fx / depth + cx
    ty = y * fy / depth + cy
    return [tx, ty, depth]


def parse_calibration(filepath: str) -> List[List[float]]:
    """Parse calibration file
    filepath: The content of a calibration file looks like this:
        Color camera intrinsic:
        0.6786797 0.90489584 0.49585155 0.5035042
        Depth camera intrinsic:
        0.6786797 0.90489584 0.49585155 0.5035042
    """
    with open(filepath, 'r') as f:
        calibration = []
        for _ in range(2):
            f.readline().strip()
            line_with_numbers = f.readline()
            intrinsic = parse_numbers(line_with_numbers)
            calibration.append(intrinsic)
    return calibration



import logging
import logging.config

import numpy as np
from depthmap import Depthmap

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d')


def export_obj(filename: str,
               dmap: Depthmap,
               triangulate: bool):
    """Export .obj file, which can be visualized in tools like Meshlab.

    triangulate=True generates OBJ of type mesh
    triangulate=False generates OBJ of type pointcloud
    """
    count = 0
    indices = np.zeros((dmap.width, dmap.height))

    # Create MTL file (a standart extension of OBJ files to define geometry materials and textures)
    material = filename[:len(filename) - 4] + '.mtl'
    if dmap.has_rgb:
        with open(material, 'w') as f:
            f.write('newmtl default\n')
            f.write('map_Kd ../' + dmap.rgb_fpath + '\n')

    with open(filename, 'w') as f:
        if dmap.has_rgb:
            f.write('mtllib ' + material[filename.index('/') + 1:] + '\n')
            f.write('usemtl default\n')
        for x in range(2, dmap.width - 2):
            for y in range(2, dmap.height - 2):
                depth = dmap.parse_depth(x, y)
                if not depth:
                    continue
                res = dmap.convert_2d_to_3d_oriented(1, x, y, depth)
                if not res:
                    continue
                count = count + 1
                indices[x][y] = count  # add index of written vertex into array
                f.write('v ' + str(res[0]) + ' ' + str(res[1]) + ' ' + str(res[2]) + '\n')
                f.write('vt ' + str(x / dmap.width) + ' ' + str(1 - y / dmap.height) + '\n')

        if triangulate:
            _do_triangulation(dmap, indices, f)
        logging.info('Mesh exported into %s', filename)


def _do_triangulation(dmap: Depthmap, indices, filehandle):
    max_diff = 0.2
    for x in range(2, dmap.width - 2):
        for y in range(2, dmap.height - 2):
            # get depth of all points of 2 potential triangles
            d00 = dmap.parse_depth(x, y)
            d10 = dmap.parse_depth(x + 1, y)
            d01 = dmap.parse_depth(x, y + 1)
            d11 = dmap.parse_depth(x + 1, y + 1)

            # check if first triangle points have existing indices
            if indices[x][y] > 0 and indices[x + 1][y] > 0 and indices[x][y + 1] > 0:
                # check if the triangle size is valid (to prevent generating triangle
                # connecting child and background)
                if abs(d00 - d10) + abs(d00 - d01) + abs(d10 - d01) < max_diff:
                    c = str(int(indices[x][y]))
                    b = str(int(indices[x + 1][y]))
                    a = str(int(indices[x][y + 1]))
                    # define triangle indices in (world coordinates / texture coordinates)
                    filehandle.write('f ' + a + '/' + a + ' ' + b + '/' + b + ' ' + c + '/' + c + '\n')

            # check if second triangle points have existing indices
            if indices[x + 1][y + 1] > 0 and indices[x + 1][y] > 0 and indices[x][y + 1] > 0:
                # check if the triangle size is valid (to prevent generating triangle
                # connecting child and background)
                if abs(d11 - d10) + abs(d11 - d01) + abs(d10 - d01) < max_diff:
                    a = str(int(indices[x + 1][y + 1]))
                    b = str(int(indices[x + 1][y]))
                    c = str(int(indices[x][y + 1]))
                    # define triangle indices in (world coordinates / texture coordinates)
                    filehandle.write('f ' + a + '/' + a + ' ' + b + '/' + b + ' ' + c + '/' + c + '\n')


def _write_pcd_header(filehandle, count):
    filehandle.write('# timestamp 1 1 float 0\n')
    filehandle.write('# .PCD v.7 - Point Cloud Data file format\n')
    filehandle.write('VERSION .7\n')
    filehandle.write('FIELDS x y z c\n')
    filehandle.write('SIZE 4 4 4 4\n')
    filehandle.write('TYPE F F F F\n')
    filehandle.write('COUNT 1 1 1 1\n')
    filehandle.write('WIDTH ' + count + '\n')
    filehandle.write('HEIGHT 1\n')
    filehandle.write('VIEWPOINT 0 0 0 1 0 0 0\n')
    filehandle.write('POINTS ' + count + '\n')
    filehandle.write('DATA ascii\n')


def export_pcd(filename: str, dmap: Depthmap):
    with open(filename, 'w') as f:
        count = str(_get_count(dmap))
        _write_pcd_header(f, count)

        for x in range(2, dmap.width - 2):
            for y in range(2, dmap.height - 2):
                depth = dmap.parse_depth(x, y)
                if not depth:
                    continue
                res = dmap.convert_2d_to_3d(1, x, y, depth)
                if not res:
                    continue
                confidence = dmap.parse_confidence(x, y)
                f.write(str(-res[0]) + ' ' + str(res[1]) + ' ' + str(res[2]) + ' ' + str(confidence) + '\n')
        logging.info('Pointcloud exported into %s', filename)


def _get_count(dmap: Depthmap) -> int:
    count = 0
    for x in range(2, dmap.width - 2):
        for y in range(2, dmap.height - 2):
            depth = dmap.parse_depth(x, y)
            if not depth:
                continue
            res = dmap.convert_2d_to_3d(1, x, y, depth)
            if not res:
                continue
            count = count + 1
    return count



from flask import Flask, request, render_template

from cgmzscore import Calculator


app = Flask(__name__)


@app.route('/')
def customer():
    file_path = '/Users/prajwalsingh/Desktop/ZScore/ZscoreWeb/image/roob_rbpp_151005.jpg'
    return render_template('first.html', user_image=file_path)


@app.route('/success', methods=['POST', 'GET'])
def print_data():
    if request.method == 'POST':
        result = request.form
        print(result['age'])
        v = Calculator().zScore_wfa(weight=str(result['weight']), muac=str(result['muca']), age_in_days=str(
            result['age']), sex=str(result['gender']), height=str(result['height']))
        print(v)
        return render_template("result.html", result=v)


if __name__ == '__main__':
    app.run(debug=True)
